import { BANK_OAUTH_CLIENT_ID, BANK_OAUTH_CLIENT_SECRET } from "@/../config.ts"
import type {
  Interval,
  OAuthTokenResponse,
  TransactionResponse,
} from "@/bank/sparebank1.ts"
import logger from "@/logger.ts"
import { toISODateString } from "@/date.ts"
import * as querystring from "node:querystring"

const baseUrl = "https://api.sparebank1.no"

type Success<T> = { status: "success"; data: T }
type Failure<T> = { status: "failure"; data: T }
type Result<OK, Err> = Success<OK> | Failure<Err>

const success = <T>(data: T): Success<T> => ({ status: "success", data: data })
const failure = <T>(data: T): Failure<T> => ({ status: "failure", data: data })

export async function transactions(
  accessToken: string,
  accountKeys: string | ReadonlyArray<string>,
  interval?: Interval,
): Promise<TransactionResponse> {
  const queryString = querystring.stringify({
    accountKey: accountKeys,
    ...(interval && {
      fromDate: toISODateString(interval.fromDate),
      toDate: toISODateString(interval.toDate),
    }),
  })

  const url = `${baseUrl}/personal/banking/transactions?${queryString}`
  logger.debug(`Sending GET request to '${url}'`)
  const response = await fetch(url, {
    headers: {
      Authorization: `Bearer ${accessToken}`,
      Accept: "application/vnd.sparebank1.v1+json;charset=utf-8",
    },
  })
  logger.debug(`Received response with status '${response.status}'`)
  if (response.ok) {
    return response.json()
  } else {
    logger.warn(await response.json())
    return { transactions: [] }
  }
}

export async function refreshToken(
  refreshToken: string,
): Promise<Result<OAuthTokenResponse, string>> {
  const queries = querystring.stringify({
    client_id: BANK_OAUTH_CLIENT_ID,
    client_secret: BANK_OAUTH_CLIENT_SECRET,
    refresh_token: refreshToken,
    grant_type: "refresh_token",
  })
  const url = `${baseUrl}/oauth/token?${queries}`
  logger.debug(`Sending POST request to url: '${url}'`)
  const response = await fetch(url, {
    method: "post",
    headers: {
      "Content-Type": "application/x-www-form-urlencoded",
    },
  })
  logger.debug(`Received response with status '${response.status}'`)
  if (!response.ok) {
    return failure(await response.text())
  }
  return success(await response.json())
}
